from __future__ import absolute_import
from ..utils import from_iso8601
import datetime
import six


class Time(object):
    """ Base of Datetime & Date
    """

    def __str__(self):
        return str(self.to_json())

    def to_json(self):
        # according to
        #   https://github.com/wordnik/swagger-spec/issues/95
        return self.v.isoformat()


class Date(Time):
    """ for string type, date format
    """

    def apply_with(self, _, v, ctx):
        """ constructor

        :param v: things used to constrcut date
        :type v: timestamp in float, datetime.date object, or ISO-8601 in str
        """
        self.v = None
        if isinstance(v, float):
            self.v = datetime.date.fromtimestamp(v)
        elif isinstance(v, datetime.date):
            self.v = v
        elif isinstance(v, six.string_types):
            self.v = from_iso8601(v).date()
        else:
            raise ValueError('Unrecognized type for Date: ' + str(type(v)))


class Datetime(Time):
    """ for string type, datetime format
    """

    def apply_with(self, _, v, ctx):
        """ constructor

        :param v: things used to constrcut date
        :type v: timestamp in float, datetime.datetime object, or ISO-8601 in str
        """
        self.v = None
        if isinstance(v, float):
            self.v = datetime.datetime.utcfromtimestamp(v)
        elif isinstance(v, datetime.datetime):
            self.v = v
        elif isinstance(v, six.string_types):
            self.v = from_iso8601(v)
        else:
            raise ValueError('Unrecognized type for Datetime: ' + str(type(v)))





from __future__ import absolute_import
from ..base import Context, ContainerType
from .objects import (
    Scope,
    LoginEndpoint,
    Implicit,
    TokenRequestEndpoint,
    TokenEndpoint,
    AuthorizationCode,
    GrantType,
    Authorization,
    Authorizations,
    ResponseMessage,
    Parameter,
    Operation,
    Api,
    Property,
    Model,
    Resource,
    Info,
    ResourceList)
from ...consts import private
from ...utils import url_dirname, url_join


class ScopeContext(Context):
    """ Context of Scope Object
    """
    __swagger_ref_object__ = Scope


class AuthorizationsContext(Context):
    """ Context of Authorization's' Object
    
    Do not get confused with Authorization Object,
    this one is used in API Declaration
    """
    __swagger_ref_object__ = Authorizations


class LoginEndpointContext(Context):
    """ Context of LoginEndpoint Object
    """
    __swagger_ref_object__ = LoginEndpoint


class ImplicitContext(Context):
    """ Context of Implicit Object
    """
    __swagger_ref_object__ = Implicit
    __swagger_child__ = {'loginEndpoint': (None, LoginEndpointContext)}


class TokenRequestEndpointContext(Context):
    """ Context of TokenRequestEndpoint Object
    """
    __swagger_ref_object__ = TokenRequestEndpoint


class TokenEndpointContext(Context):
    """ Context of Token Object
    """
    __swagger_ref_object__ = TokenEndpoint


class AuthorizationCodeContext(Context):
    """ Context of AuthorizationCode Object
    """
    __swagger_ref_object__ = AuthorizationCode
    __swagger_child__ = {
        'tokenRequestEndpoint': (None, TokenRequestEndpointContext),
        'tokenEndpoint': (None, TokenEndpointContext)
    }


class GrantTypeContext(Context):
    """ Context of GrantType Object
    """
    __swagger_ref_object__ = GrantType
    __swagger_child__ = {
        'implicit': (None, ImplicitContext),
        'authorization_code': (None, AuthorizationCodeContext)
    }


class AuthorizationContext(Context):
    """ Context of Authorization Object
    """
    __swagger_ref_object__ = Authorization
    __swagger_child__ = {
        'scopes': (ContainerType.list_, ScopeContext),
        'grantTypes': (None, GrantTypeContext)
    }


class ResponseMessageContext(Context):
    """ Context of ResponseMessage Object
    """
    __swagger_ref_object__ = ResponseMessage
 
class ParameterContext(Context):
    """ Context of Parameter Object
    """
    __swagger_ref_object__ = Parameter 

class OperationContext(Context):
    """ Context of Operation Object
    """
    __swagger_ref_object__ = Operation
    __swagger_child__ = {
        'authorizations': (ContainerType.dict_of_list_, AuthorizationsContext),
        'parameters': (ContainerType.list_, ParameterContext),
        'responseMessages': (ContainerType.list_, ResponseMessageContext)
    }


class ApiContext(Context):
    """ Context of Api Object
    """
    __swagger_ref_object__ = Api
    __swagger_child__ = {'operations': (ContainerType.list_, OperationContext)}


class PropertyContext(Context):
    """ Context of Property Object
    """
    __swagger_ref_object__ = Property


class ModelContext(Context):
    """ Context of Model Object
    """
    __swagger_ref_object__ = Model
    __swagger_child__ = {'properties': (ContainerType.dict_, PropertyContext)}

class ResourceContext(Context):
    """ Context of Resource Object
    """
    __swagger_ref_object__ = Resource
    __swagger_child__ = {
        'apis': (ContainerType.list_, ApiContext),
        'models': (ContainerType.dict_, ModelContext)
    }


class InfoContext(Context):
    """ Context of Info Object
    """
    __swagger_ref_object__ = Info


class ResourceListContext(Context):
    """ Context of Resource List Object
    """
    __swagger_ref_object__ = ResourceList
    __swagger_child__ = {
        'info': (None, InfoContext),
        'authorizations': (ContainerType.dict_, AuthorizationContext)
    }

    def __init__(self, parent, backref):
        super(ResourceListContext, self).__init__(parent, backref)

    def parse(self, obj, root_url, resolver, getter):
        super(ResourceListContext, self).parse(obj=obj)

        resources = []
        if private.SCHEMA_APIS in obj:
            if isinstance(obj[private.SCHEMA_APIS], list):
                for api in obj[private.SCHEMA_APIS]:
                    resources.append(api[private.SCHEMA_PATH])
            else:
                raise TypeError('Invalid type of apis: ' + type(obj[private.SCHEMA_APIS]))
        base = url_dirname(root_url)
        urls = zip(
            map(lambda u: url_join(base,  u[1:]), resources),
            map(lambda u: u[1:], resources)
        )

        # replace each element in 'apis' with Resource
        self._obj['apis'] = {}
        # get into resource object
        for url, name in urls:
            # here we assume Resource is always a dict
            self._obj['apis'][name] = {}
            res = resolver.resolve(url, getter)
            with ResourceContext(self._obj['apis'], name) as ctx:
                ctx.parse(obj=res)




from __future__ import absolute_import
from .consts import private
from .utils import patch_path
import json
import yaml
import six
import os
import logging
import re


logger = logging.getLogger(__name__)


class Getter(six.Iterator):
    """ base of getter object

    Idealy, to subclass a getter, you just need to override load function.
    The part to extend getter would be finalized once Swagger 2.0 is ready.
    """

    def __init__(self, path):
        self.base_path = path

    def __iter__(self):
        return self

    def __next__(self):
        if len(self.urls) == 0:
            raise StopIteration

        obj = self.load(self.urls.pop(0))

        # make sure data is string type
        if isinstance(obj, dict):
            pass
        elif isinstance(obj, six.binary_type):
            obj = obj.decode('utf-8')
        elif not isinstance(obj, six.string_types):
            raise ValueError('Unknown types: [{0}]'.format(str(type(obj))))

        # a very simple logic to distinguish json and yaml
        if isinstance(obj, six.string_types):
            try:
                if obj.startswith('{'):
                    obj = json.loads(obj)
                else:
                    obj = yaml.safe_load(obj)
            except ValueError:
                raise Exception('Unknown format startswith {0} ...'.format(obj[:10]))

        return obj

    def load(self, path):
        """ load the resource, and return for parsing.

        :return: name and json object of resources
        :rtype: (str, dict)
        """
        raise NotImplementedError()

class LocalGetter(Getter):
    """ default getter implmenetation for local resource file
    """
    def __init__(self, path):
        super(LocalGetter, self).__init__(path)

        if path.startswith('file://'):
            parsed = six.moves.urllib.parse.urlparse(path)
            path = parsed.path
        if re.match('^/[A-Z]+:', path) is not None:
            path = os.path.abspath(path[1:])

        for n in private.SWAGGER_FILE_NAMES:
            if self.base_path.endswith(n):
                self.base_path = os.path.dirname(self.base_path)
                self.urls = [path]
                break
            else:
                p = os.path.join(path, n)
                if os.path.isfile(p):
                    self.urls = [p]
                    break
        else:
            # there is no file matched predefined file name:
            # - resource_list.json (1.2)
            # - swagger.json       (2.0)
            # in this case, we will locate them in this way:
            # - when 'path' points to a specific file, and its
            #   extension is either 'json' or 'yaml'.
            _, ext = os.path.splitext(path)
            for e in [private.FILE_EXT_JSON, private.FILE_EXT_YAML, private.FILE_EXT_YML]:
                if ext.endswith(e):
                    self.base_path = os.path.dirname(path)
                    self.urls = [path]
                    break
            else:
                for e in [private.FILE_EXT_JSON, private.FILE_EXT_YAML, private.FILE_EXT_YML]:
                    #print(path + '.' + e)
                    if os.path.isfile(path + '.' + e):
                        self.urls = [path + '.' + e]
                        break
                else:
                    raise ValueError('Unable to locate resource file: [{0}]'.format(path))

    def load(self, path):
        logger.info('to load: [{0}]'.format(path))

        path = patch_path(self.base_path, path)
        logger.info('final path to load: [{0}]'.format(path))

        ret = None
        with open(path, 'r') as f:
            ret = f.read()
        return ret


class SimpleGetter(Getter):
    """ the simple getter that don't have to concern file loading of LocalGetter
    """

    __simple_getter_callback__ = lambda url: {}
    """ the callback to load the resource, accept an URL and return a string buffer
    """

    def __init__(self, path):
        if isinstance(path, six.string_types):
            super(SimpleGetter, self).__init__(path)
            if self.base_path.endswith('/'):
                self.base_path = self.base_path[:-1]
            self.urls = [path]
        else:
            raise Exception('Unsupported type for "path": {} in SimpleGetter'.format(str(type(path))))

    def load(self, path):
        logger.info('to load: [{0}]'.format(path))

        return self.__simple_getter_callback__.__func__(path)


def _url_load(path):
    ret = f = None
    try:
        f = six.moves.urllib.request.urlopen(path)
        ret = f.read()
    finally:
        if f:
            f.close()

    return ret

class UrlGetter(SimpleGetter):
    """ default getter implementation for remote resource file
    """
    __simple_getter_callback__ = _url_load


class DictGetter(Getter):
    """ a getter accept a dict as parameter without loading from file / url

    args:
     - urls: the urls to be loaded in upcoming resolving (the order should be matched to get result correct)
     - path2dict: a mapping from 'path' to 'dict', which is the mocking of 'downloaded data'
    """
    def __init__(self, urls, path2dict):
        super(DictGetter, self).__init__(urls[0])
        self.urls = urls
        self._path2dict = path2dict or {}

    def load(self, path):
        logger.info('to load: [{0}]'.format(path))

        return self._path2dict.get(path, {})



from __future__ import absolute_import
from ...scan import Dispatcher
from ...spec.v2_0.objects import (
    Items,
    Schema,
    Parameter,
    Header,
    Response,
    Operation,
    PathItem,
    Swagger,
    )
from ...spec.v2_0.parser import (
    ItemsContext,
    SchemaContext,
    ParameterContext,
    HeaderContext,
    ResponseContext,
    OperationContext,
    PathItemContext,
    SwaggerContext,
    )


class AssignParent(object):
    """ parent assignment """

    class Disp(Dispatcher): pass

    @Disp.register([Items])
    def _items(self, path, obj, _):
        obj._assign_parent(ItemsContext)

    @Disp.register([Schema])
    def _schema(self, path, obj, _):
        obj._assign_parent(SchemaContext)

    @Disp.register([Parameter])
    def _parameter(self, path, obj, _):
        obj._assign_parent(ParameterContext)

    @Disp.register([Header])
    def _header(self, path, obj, _):
        obj._assign_parent(HeaderContext)

    @Disp.register([Response])
    def _response(self, path, obj, _):
        obj._assign_parent(ResponseContext)

    @Disp.register([Operation])
    def _operation(self, path, obj, _):
        obj._assign_parent(OperationContext)

    @Disp.register([PathItem])
    def _path_item(self, path, obj, _):
        obj._assign_parent(PathItemContext) 

    @Disp.register([Swagger])
    def _swagger(self, path, obj, _):
        obj._assign_parent(SwaggerContext)
 



from __future__ import absolute_import
from ..base import BaseObj, FieldMeta
from ...utils import final
from ...io import Request
from ...io import Response as _Response
from ...primitives import Array
import six
import copy


class BaseObj_v2_0(BaseObj):
    __swagger_version__ = '2.0'


class XMLObject(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ XML Object
    """
    __swagger_fields__ = {
        'name': None,
        'namespace': None,
        'prefix': None,
        'attribute': None,
        'wrapped': None,
    }


class BaseSchema(BaseObj_v2_0):
    """ Base type for Items, Schema, Parameter, Header
    """

    __swagger_fields__ = {
        'type': None,
        'format': None,
        'items': None,
        'default': None,
        'maximum': None,
        'exclusiveMaximum': None,
        'minimum': None,
        'exclusiveMinimum': None,
        'maxLength': None,
        'minLength': None,
        'maxItems': None,
        'minItems': None,
        'multipleOf': None,
        'enum': None,
        'pattern': None,
        'uniqueItems': None,
    }


class Items(six.with_metaclass(FieldMeta, BaseSchema)):
    """ Items Object
    """

    __swagger_fields__ = {
        'collectionFormat': 'csv',
    }

    def _prim_(self, v, prim_factory, ctx=None):
        return prim_factory.produce(self, v, ctx)


class Schema(six.with_metaclass(FieldMeta, BaseSchema)):
    """ Schema Object
    """

    __swagger_fields__ = {
        '$ref': None,
        'maxProperties': None,
        'minProperties': None,
        'required': [],
        'allOf': [],
        'properties': {},
        'additionalProperties': True,
        'title': None,
        'description': None,
        'discriminator': None,
        'readOnly': None,
        'xml': None,
        'externalDocs': None,
        'example': None,
    }

    __internal_fields__ = {
        # pyswagg only
        'ref_obj': None,
        'final': None,
        'name': None,
    }

    def _prim_(self, v, prim_factory, ctx=None):
        return prim_factory.produce(self, v, ctx)


class Swagger(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ Swagger Object
    """

    __swagger_fields__ = {
        'swagger': None,
        'info': None,
        'host': None,
        'basePath': None,
        'schemes': [],
        'consumes': [],
        'produces': [],
        'paths': None,
        'definitions': None,
        'parameters': None,
        'responses': None,
        'securityDefinitions': None,
        'security': None,
        'tags': None,
        'externalDocs': None,
    }


class Contact(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ Contact Object
    """

    __swagger_fields__ = {
        'name': None,
        'url': None,
        'email': None,
    }


class License(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ License Object
    """

    __swagger_fields__ = {
        'name': None,
        'url': None,
    }


class Info(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ Info Object
    """

    __swagger_fields__ = {
        'version': None,
        'title': None,
        'description': None,
        'termsOfService': None,
        'contact': None,
        'license': None,
    }


class Parameter(six.with_metaclass(FieldMeta, BaseSchema)):
    """ Parameter Object
    """

    __swagger_fields__ = {
        # Reference Object
        '$ref': None,

        'name': None,
        'in': None,
        'required': None,

        # body parameter
        'schema': None,

        # other parameter
        'collectionFormat': 'csv',

        # for converter only
        'description': None,

        # TODO: not supported yet
        'allowEmptyValue': False,
    }

    __internal_fields__ = {
        'final': None,
    }

    def _prim_(self, v, prim_factory, ctx=None):
        i = getattr(self, 'in')
        return prim_factory.produce(self.schema, v, ctx) if i == 'body' else prim_factory.produce(self, v, ctx)


class Header(six.with_metaclass(FieldMeta, BaseSchema)):
    """ Header Object
    """

    __swagger_fields__ = {
        'collectionFormat': 'csv',
        'description': None,
    }

    def _prim_(self, v, prim_factory, ctx=None):
        return prim_factory.produce(self, v, ctx)


class Response(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ Response Object
    """

    __swagger_fields__ = {
        # Reference Object
        '$ref': None,

        'schema': None,
        'headers': {},

        'description': None,
        'examples': None,
    }

    __internal_fields__ = {
        'final': None,
    }


class Operation(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ Operation Object
    """

    __swagger_fields__ = {
        'tags': None,
        'operationId': None,
        'consumes': [],
        'produces': [],
        'schemes': [],
        'parameters': None,
        'responses': None,
        'deprecated': False,
        'security': None,
        'description': None,
        'summary': None,
        'externalDocs': None,
    }

    __internal_fields__ = {
        'method': None,
        'url': None,
        'path': None,
        'base_path': None,
        'cached_schemes': [],
    }

    def __call__(self, **k):
        # prepare parameter set
        params = dict(header={}, query=[], path={}, body={}, formData=[], file={})
        names = []
        def _convert_parameter(p):
            if p.name not in k and not p.is_set("default") and p.required:
                raise ValueError('requires parameter: ' + p.name)

            if p.is_set("default"):
                v = k.get(p.name, p.default)
            else:
                if p.name in k:
                    v = k[p.name]
                else:
                    # do not provide value for parameters that use didn't specify.
                    return

            c = p._prim_(v, self._prim_factory, ctx=dict(read=False))
            i = getattr(p, 'in')

            if p.type == 'file':
                params['file'][p.name] = c
            elif i in ('query', 'formData'):
                if isinstance(c, Array):
                    if p.items.type == 'file':
                        params['file'][p.name] = c
                    else:
                        params[i].extend([tuple([p.name, v]) for v in c.to_url()])
                else:
                    params[i].append((p.name, str(c),))
            else:
                params[i][p.name] = str(c) if i != 'body' else c

            names.append(p.name)

        for p in self.parameters:
            _convert_parameter(final(p))

        # check for unknown parameter
        unknown = set(six.iterkeys(k)) - set(names)
        if len(unknown) > 0:
            raise ValueError('Unknown parameters: {0}'.format(unknown))

        return \
        Request(op=self, params=params), _Response(self)


class PathItem(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ Path Item Object
    """

    __swagger_fields__ = {
        # Reference Object
        '$ref': None,

        'get': None,
        'put': None,
        'post': None,
        'delete': None,
        'options': None,
        'head': None,
        'patch': None,
        'parameters': [],
    }


class SecurityScheme(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ Security Scheme Object
    """

    __swagger_fields__ = {
        'type': None,
        'name': None,
        'in': None,
        'flow': None,
        'authorizationUrl': None,
        'tokenUrl': None,
        'scopes': None,
        'description': None,
    }


class Tag(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ Tag Object
    """

    __swagger_fields__ = {
        'name': None,
        'description': None,
        'externalDocs': None,
    }


class ExternalDocumentation(six.with_metaclass(FieldMeta, BaseObj_v2_0)):
    """ External Documentation Object
    """

    __swagger_fields__ = {
        'description': None,
        'url': None,
    }



from __future__ import absolute_import
from ..spec.v2_0.objects import Parameter, Operation, Schema
from ..utils import deref, final, from_iso8601
from decimal import Decimal
import random
import six
import sys
import string
import uuid
import base64
import datetime
import time

# TODO: patternProperties
# TODO: pattern
# TODO: binary
# TODO: enum of object, array

# min/max of integer
maxInt32 = 1 << 31 - 1
minInt32 = -maxInt32
maxInt64 = 1 << 63 - 1
minInt64 = -maxInt64

def _int_(obj, _, val=None):
    if val:
        return val

    max_ = maxInt32 if getattr(obj, 'format') in ['int32', None] else maxInt64
    max_ = obj.maximum if obj.maximum else max_
    max_ = max_-1 if obj.exclusiveMaximum else max_

    min_ = minInt32 if getattr(obj, 'format') in ['int32', None] else minInt64
    min_ = obj.minimum if obj.minimum else min_
    min_ = min_+1 if obj.exclusiveMinimum else min_

    out = random.randint(min_, max_)
    return out - (out % obj.multipleOf) if isinstance(obj.multipleOf, six.integer_types) and obj.multipleOf != 0 else out

def _float_(obj, _, val=None):
    if val:
        return val

    # TODO: exclusiveMaximum == False is not implemented.
    max_ = obj.maximum if obj.maximum else sys.float_info.max
    min_ = obj.minimum if obj.minimum else sys.float_info.min

    out = None
    while out == None:
        out = min_ + (max_ - min_) * random.random()
        if obj.multipleOf and obj.multipleOf != 0:
            out = int(out / obj.multipleOf) * obj.multipleOf
        if out <= min_ and obj.exclusiveMinimum:
            out = None
    return float(out)

def _str_(obj, opt, val=None):
    if val:
        return val

    max_ = obj.maxLength if obj.maxLength else opt['max_str_length']
    min_ = obj.minLength if obj.minLength else 0

    # note: length is 0~100, characters are limited to ASCII
    return ''.join([random.choice(string.ascii_letters) for _ in range(random.randint(min_, max_))])

def _bool_(obj, _, val=None):
    return bool(val) if val else random.randint(0, 1) == 0

def _uuid_(obj, _, val=None):
    return uuid.UUID(val) if val else uuid.uuid4()

names = list(string.ascii_letters) + ['_', '-'] + list(string.digits)
def _email_name_():
    return random.choice(string.ascii_letters) \
    + ''.join([random.choice(names) for _ in six.moves.xrange(random.randint(1, 30))]) \
    + random.choice(string.ascii_letters)

def _email_(obj, _, val=None):
    if val:
        return val

    host_length = random.randint(2, 100)
    region_length = random.randint(2, 30)
    return '.'.join([_email_name_() for _ in six.moves.xrange(random.randint(1, 4))]) \
        + '@' \
        + random.choice(string.ascii_letters) \
        + ''.join([random.choice(names) for _ in six.moves.xrange(host_length)]) \
        + '.' \
        + random.choice(string.ascii_letters) \
        + ''.join([random.choice(names) for _ in six.moves.xrange(region_length)])

def _byte_(obj, opt, val=None):
    return val if val else base64.b64encode(
        six.b(''.join([random.choice(string.ascii_letters) for _ in range(random.randint(0, opt['max_byte_length']))]))
    )

max_date = time.mktime(datetime.date(2038, 1, 19).timetuple())
min_date = time.mktime(datetime.date(1970, 1, 2).timetuple())
def _date_(obj, _, val=None):
    return from_iso8601(val).date() if val else datetime.date.fromtimestamp(
        random.uniform(min_date, max_date)
    )

max_datetime = time.mktime(datetime.datetime(2038, 1, 19).utctimetuple())
min_datetime = time.mktime(datetime.datetime(1970, 1, 2).utctimetuple())
def _date_time_(obj, _, val=None):
    return from_iso8601(val) if val else datetime.datetime.utcfromtimestamp(
        random.uniform(min_datetime, max_datetime)
    )

def _file_(obj, opt, _):
    if len(opt['files'] or []) > 0:
        return random.choice(opt['files'])
    return dict(
        header={
            'Content-Type': 'text/plain',
            'Content-Transfer-Encoding': 'binary'
        },
        filename='',
        data=six.moves.cStringIO(
            ''.join([random.choice(string.ascii_letters) for _ in range(random.randint(0, opt['max_file_length']))])
        )
    )

class Renderer(object):
    """
    """

    def __init__(self):
        """
        """
        random.seed()

        # init map of generators
        self._map = {
            'integer': {
                '': _int_,
                None: _int_,
                'int32': _int_,
                'int64': _int_,
            },
            'number': {
                '': _float_,
                None: _float_,
                'float': _float_,
                'double': _float_,
            },
            'string': {
                '': _str_,
                None: _str_,
                'password': _str_,
                'byte': _byte_,
                'date': _date_,
                'date-time': _date_time_,
                'uuid': _uuid_,
                'email': _email_,
            },
            'boolean': {
                '': _bool_,
                None: _bool_,
            },
            'file': {
                '': _file_,
                None: _file_,
            }
        }

    def _get(self, _type, _format=None):
        r = self._map.get(_type, None)
        return None if r == None else r.get(_format, None)

    def _generate(self, obj, opt):
        obj = final(deref(obj))
        type_ = getattr(obj, 'type', None)
        template = opt['object_template']
        out = None
        if type_ == 'object':
            max_p = opt['max_property']
            out = {}
            max_ = obj.maxProperties if obj.maxProperties else opt['max_prop_count']
            min_ = obj.minProperties if obj.minProperties else None
            count = 0
            for name, prop in six.iteritems(obj.properties or {}):
                if name in template:
                    out[name] = template[name]
                    continue
                if not max_p and not name in obj.required:
                    if random.randint(0, 1) == 0 or opt['minimal_property']:
                        continue
                out[name] = self._generate(prop, opt)
                count = count + 1

            if isinstance(obj.additionalProperties, Schema):
                # TODO: additionalProperties == True is not handled

                # generate random properties
                more = random.randint(min_, max_) - count
                if more > 0:
                    # generate a random string as property-name
                    for _ in six.moves.xrange(more):
                        while True:
                            length = random.randint(0, opt['max_name_length'])
                            name = ''.join([random.choice(string.ascii_letters) for _ in six.moves.xrange(length)])
                            if name not in out:
                                out[name] = self._generate(obj.additionalProperties, opt)
                                break

        elif type_ == 'array':
            min_ = obj.minItems if obj.minItems else 0
            max_ = obj.maxItems if obj.maxItems else opt['max_array_length']
            out = []
            for _ in six.moves.xrange(random.randint(min_, max_)):
                out.append(self._generate(obj.items, opt))

        elif type_ != None:
            out = None
            if len(obj.enum or []) > 0:
                out = random.choice(obj.enum)

            g = self._get(getattr(obj, 'type', None), getattr(obj, 'format', None))
            if not g:
                raise Exception('Unable to locate generator: {0}'.format(obj))
            out = g(obj, opt, out)
        else:
            raise Exception('No type info available:{0}, {1}'.format(obj.type, obj.format))

        return out

    @staticmethod
    def default():
        """ return default options, available options:
        - max_name_length: maximum length of name for additionalProperties
        - max_prop_count: maximum count of properties (count of fixed properties + additional properties)
        - max_str_length: maximum length of string type
        - max_byte_length: maximum length of byte type
        - max_array_length: maximum length of array
        - max_file_length: maximum length of file, in byte
        - minimal_property: only generate 'required' properties
        - minimal_parameter: only generate 'required' parameter
        - files: registered file object: refer to pyswagg.primitives.File for details
        - object_template: dict of default values assigned for properties when 'name' matched
        - parameter_template: dict of default values assigned for parameters when 'name matched
        - max_property: all properties are generated, ignore 'required'
        - max_parameter: all parameters are generated, ignore 'required'

        :return: options
        :rtype: dict
        """
        return dict(
            max_name_length=64,
            max_prop_count=32,
            max_str_length=50,
            max_byte_length=50,
            max_array_length=10,
            max_file_length=100,
            minimal_property=False,
            minimal_parameter=False,
            files=[],
            object_template={},
            parameter_template={},
            max_property=False,
            max_parameter=False,
        )

    def render(self, obj, opt=None):
        """ render a Schema/Parameter

        :param obj Schema/Parameter: the swagger spec object
        :param opt dict: render option
        :return: values that can be passed to Operation.__call__
        :rtype: depends on type of 'obj'
        """
        opt = self.default() if opt == None else opt
        if not isinstance(opt, dict):
            raise ValueError('Not a dict: {0}'.format(opt))

        if isinstance(obj, Parameter):
            if getattr(obj, 'in', None) == 'body':
                return self._generate(obj.schema, opt)
            return self._generate(obj, opt=opt)
        elif isinstance(obj, Schema):
            return self._generate(obj, opt)
        else:
            raise ValueError('Not a Schema/Parameter: {0}'.format(obj))

    def render_all(self, op, exclude=[], opt=None):
        """ render a set of parameter for an Operation

        :param op Operation: the swagger spec object
        :param opt dict: render option
        :return: a set of parameters that can be passed to Operation.__call__
        :rtype: dict
        """
        opt = self.default() if opt == None else opt
        if not isinstance(op, Operation):
            raise ValueError('Not a Operation: {0}'.format(op))
        if not isinstance(opt, dict):
            raise ValueError('Not a dict: {0}'.format(opt))

        template = opt['parameter_template']
        max_p = opt['max_parameter']
        out = {}
        for p in op.parameters:
            if p.name in exclude:
                continue
            if p.name in template:
                out.update({p.name: template[p.name]})
                continue
            if not max_p and not p.required:
                if random.randint(0, 1) == 0 or opt['minimal_parameter']:
                    continue
            out.update({p.name: self.render(p, opt=opt)})
        return out



from __future__ import absolute_import
from ..utils import walk
from ..scan import Dispatcher
from ..spec.v2_0.objects import (
    Schema,
    Parameter,
    Response,
    PathItem,
    )
from ..spec.v2_0.parser import (
    SchemaContext,
    ParameterContext,
    ResponseContext,
    PathItemContext,
    )
import functools
import six

def _out(app, parser, path):
    obj = app.resolve(path, parser=parser)
    r = getattr(obj, '$ref')
    return [r] if r else []

def _schema_out_obj(obj, out=None):
    out = [] if out == None else out

    for o in six.itervalues(obj.properties):
        out = _schema_out_obj(o, out)

    for o in obj.allOf:
        out = _schema_out_obj(o, out)

    if isinstance(obj.additionalProperties, Schema):
        out = _schema_out_obj(obj.additionalProperties, out)

    if obj.items:
        out = _schema_out_obj(obj.items, out)

    r = getattr(obj, '$ref')
    if r:
        out.append(r)

    return out

def _schema_out(app, path):
    obj = app.resolve(path, parser=SchemaContext)
    return [] if obj == None else _schema_out_obj(obj)


class CycleDetector(object):
    """ circular detector """

    class Disp(Dispatcher): pass

    def __init__(self):
        self.cycles = {
            'schema':[],
            'parameter':[],
            'response':[],
            'path_item':[]
        }

    @Disp.register([Schema])
    def _schema(self, path, _, app):
        self.cycles['schema'] = walk(
            path,
            functools.partial(_schema_out, app),
            self.cycles['schema']
        )

    @Disp.register([Parameter])
    def _parameter(self, path, _, app):
         self.cycles['parameter'] = walk(
            path,
            functools.partial(_out, app, ParameterContext),
            self.cycles['parameter']
        )

    @Disp.register([Response])
    def _response(self, path, _, app):
        self.cycles['response'] = walk(
            path,
            functools.partial(_out, app, ResponseContext),
            self.cycles['response']
        )

    @Disp.register([PathItem])
    def _path_item(self, path, _, app):
        self.cycles['path_item'] = walk(
            path,
            functools.partial(_out, app, PathItemContext),
            self.cycles['path_item']
        )




from __future__ import absolute_import
from ...errs import CycleDetectionError
from ...scan import Dispatcher
from ...spec.v2_0.parser import (
    ParameterContext,
    ResponseContext,
    PathItemContext
    )
from ...spec.v2_0.objects import (
    Parameter,
    Response,
    PathItem,
    )
from ...spec.base import NullContext
from ...utils import CycleGuard


def _merge(obj, app, creator, parser):
    """ resolve $ref, and inject/merge referenced object to self.
    This operation should be carried in a cascade manner.
    """
    result = creator(NullContext())
    result.merge(obj, parser)

    guard = CycleGuard()
    guard.update(obj)

    r = getattr(obj, '$ref')
    while r and len(r) > 0:
        ro = app.resolve(r, parser)
        if ro.__class__ != obj.__class__:
            raise TypeError('Referenced Type mismatch: {0}'.format(r))
        try:
            guard.update(ro)
        except CycleDetectionError:
            # avoid infinite loop,
            # cycle detection has a dedicated scanner.
            break

        result.merge(ro, parser)
        r = getattr(ro, '$ref')
    return result

class Merge(object):
    """ pre-merge these objects with '$ref' """

    class Disp(Dispatcher): pass

    @Disp.register([Parameter])
    def _parameter(self, _, obj, app):
        obj.update_field('final', _merge(obj, app, Parameter, ParameterContext))

    @Disp.register([Response])
    def _response(self, _, obj, app):
        obj.update_field('final', _merge(obj, app, Response, ResponseContext))

    @Disp.register([PathItem])
    def _path_item(self, _, obj, app):
        obj.merge(_merge(obj, app, PathItem, PathItemContext), PathItemContext)




from __future__ import absolute_import
from ...core import BaseClient
import webapp2
import six
import sys


class Webapp2TestClient(BaseClient):
    """ Webapp2 Test Client
    """

    __schemes__ = set(['http', 'https'])

    def __init__(self, app, auth=None, keep_cookie=False, **kw):
        """ constructor

        :param app webapp2.WSGIApplication: the WSGI application to be tested
        :param auth pyswagg.Security: auth info holder
        :param keep_cookie bool: keep cookie as session or not
        """
        super(Webapp2TestClient, self).__init__(auth)

        #
        # if webapp2 finally support python 3, please submit an issue to remove this check
        #
        if sys.version_info[0] > 2:
            raise Exception('webapp2 only support python 2.x')

        self.__app = app
        self.__keep_cookie = keep_cookie
        self.__cookie = None
        self.__kw = kw

    def request(self, req_and_resp, opt=None, headers=None):
        """
        """

        # make sure all prepared state are clean before processing
        req, resp = req_and_resp
        req.reset()
        resp.reset()

        opt = opt or {}
        req, resp = super(Webapp2TestClient, self).request((req, resp), opt)

        req.prepare(scheme=self.prepare_schemes(req), handle_files=True)
        req._patch(opt)

        composed_headers = self.compose_headers(req, headers, opt)

        url = req.url
        if req.query:
            if url[-1] not in ('?', '&'):
                url += '&' if ('?' in url) else '?'
            url += six.moves.urllib.parse.urlencode(req.query)

        # initiate an request every time
        _req = webapp2.Request.blank(
            url,
            headers=composed_headers,
            POST=req.data,
            **self.__kw
        )
        _req.method = req.method.upper()
        if self.__cookie:
            _req.headers['Cookie'] = self.__cookie

        # perform the request
        _resp = _req.get_response(self.__app)

        # prepare response
        resp.apply_with(
            status=_resp.status_int,
            raw=_resp.body,
            header=_resp.headerlist,
            )

        if self.__keep_cookie:
            if 'Set-Cookie' in _resp.headers:
                self.__cookie = _resp.headers['Set-Cookie']

        return resp




from setuptools import setup, find_packages

version = open('pyswagg/__init__.py', 'r').readline().split()[2].strip("'")

setup(
    name = 'pyswagg',
    packages = find_packages(exclude=['*.tests.*']),
    version = version,
    description = 'A type-safe, dynamic, spec-compliant swagger client & converter for python',
    author = 'Mission Liao / benkilimnik',
    author_email = 'missionaryliao@gmail.com',
    url = 'https://github.com/benkilimnik/pyswagg', # use the URL to the github repo
    download_url = 'https://github.com/benkilimnik/pyswagg/tarball/{0}'.format(version),
    keywords = ['swagger', 'REST'], # arbitrary keywords
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires = ['six >= 1.7.2', 'pyaml>=15.03.1', 'validate_email']
)




from __future__ import absolute_import
from .consts import private
from .errs import CycleDetectionError
import six
import imp
import sys
import datetime
import re
import os
import operator
import functools
import collections

#TODO: accept varg
def scope_compose(scope, name, sep=private.SCOPE_SEPARATOR):
    """ compose a new scope

    :param str scope: current scope
    :param str name: name of next level in scope
    :return the composed scope
    """

    if name == None:
        new_scope = scope
    else:
        new_scope = scope if scope else name

    if scope and name:
        new_scope = scope + sep + name

    return new_scope

def scope_split(scope, sep=private.SCOPE_SEPARATOR):
    """ split a scope into names

    :param str scope: scope to be splitted
    :return: list of str for scope names
    """

    return scope.split(sep) if scope else [None]


class ScopeDict(dict):
    """ ScopeDict
    """
    def __init__(self, *a, **k):
        self.__sep = private.SCOPE_SEPARATOR
        super(ScopeDict, self).__init__(*a, **k)

    @property
    def sep(self):
        """ separator property
        """
        raise TypeError('sep property is write-only')

    @sep.setter
    def sep(self, sep):
        """ update separater used here
        """
        self.__sep = sep

    def __getitem__(self, *keys):
        """ to access an obj with key: 'n!##!m...!##!z', caller can pass as key:
        - n!##!m...!##!z
        - n, m, ..., z
        - z
        when separater == !##!

        :param dict keys: keys to access via scopes.
        """
        k = six.moves.reduce(lambda k1, k2: scope_compose(k1, k2, sep=self.__sep), keys[0]) if isinstance(keys[0], tuple) else keys[0]
        try:
            return super(ScopeDict, self).__getitem__(k)
        except KeyError as e:
            ret = []
            for ik in self.keys():
                if ik.endswith(k):
                    ret.append(ik)
            if len(ret) == 1:
                return super(ScopeDict, self).__getitem__(ret[0])
            elif len(ret) > 1:
                # special case for the last token:
                #  - a!##!get
                #  - b!##!something-get
                #  and access with 'get'
                last_k = k.rsplit(self.__sep, 1)[-1]
                matched = [r for r in ret if r.rsplit(self.__sep, 1)[-1] == last_k]
                if len(matched) == 1:
                    return super(ScopeDict, self).__getitem__(matched[0])

                raise ValueError('Multiple occurrence of key: {0}'.format(k))

            raise e


class CycleGuard(object):
    """ Guard for cycle detection
    """

    def __init__(self):
        self.__visited = []

    def update(self, obj):
        if obj in self.__visited:
            raise CycleDetectionError('Cycle detected: {0}'.format(getattr(obj, '$ref', None)))
        self.__visited.append(obj)


# TODO: this function and datetime don't handle leap-second.
#       check if dateutil handle it or not

class FixedTZ(datetime.tzinfo):
    """ tzinfo implementation without consideration of
    daylight-saving-time.
    """

    def __init__(self, h=0, m=0):
        self.__offset = datetime.timedelta(hours=h, minutes=m)

    def utcoffset(self, dt):
        return self.__offset + self.dst(dt)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return datetime.timedelta(0)

_iso8601_fmt = re.compile(''.join([
    '(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})', # YYYY-MM-DD
    'T', # T
    '(?P<hour>\d{2}):(?P<minute>\d{2})(:(?P<second>\d{1,2})(\.(?P<microsecond>\d{1,6}))?)?', # hh:mm:ss.ms
    '(?P<tz>Z|[+-]\d{2}:\d{2})?' # Z or +/-hh:mm
]))
_iso8601_fmt_date = re.compile('(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})') # YYYY-MM-DD

def from_iso8601(s):
    """ convert iso8601 string to datetime object.
    refer to http://xml2rfc.ietf.org/public/rfc/html/rfc3339.html#anchor14
    for details.

    :param str s: time in ISO-8601
    :rtype: datetime.datetime
    """
    m = _iso8601_fmt.match(s)
    if not m:
        m = _iso8601_fmt_date.match(s)

    if not m:
        raise ValueError('not a valid iso 8601 format string:[{0}]'.format(s))

    g = m.groupdict()

    def _default_zero(key):
        v = g.get(key, None)
        return int(v) if v else 0

    def _default_none(key):
        v = g.get(key, None)
        return int(v) if v else None

    year = _default_zero('year')
    month = _default_zero('month')
    day = _default_zero('day')
    hour = _default_none('hour')
    minute = _default_none('minute')
    second = _default_none('second')
    microsecond = g.get('microsecond', None)
    if microsecond is not None:
        # append zero when not matching 6 digits
        microsecond = int(microsecond + '0' * (6 - len(microsecond)))
    tz_s = g.get('tz')

    if not (year and month and day):
        raise ValueError('missing y-m-d: [{0}]'.format(s))

    # only date part, time part is none
    if hour == None and minute == None and second == None:
        return datetime.datetime(year, month, day)

    # prepare tz info
    tz = None
    if tz_s:
        if hour is None and minute is None:
            raise ValueError('missing h:m when tzinfo is provided: [{0}]'.format(s))

        negtive = hh = mm = 0

        if tz_s != 'Z':
            negtive = -1 if tz_s[0] == '-' else 1
            hh = int(tz_s[1:3])
            mm = int(tz_s[4:6]) if len(tz_s) > 5 else 0

        tz = FixedTZ(h=hh*negtive, m=mm*negtive)

    return datetime.datetime(
        year=year,
        month=month,
        day=day,
        hour=hour or 0,
        minute=minute or 0,
        second=second or 0,
        microsecond=microsecond or 0,
        tzinfo=tz
    )

def import_string(name):
    """ import module
    """
    mod = fp = None

    # code below, please refer to
    #   https://docs.python.org/2/library/imp.html
    # for details
    try:
        return sys.modules[name]
    except KeyError:
        pass

    try:
        fp, pathname, desc = imp.find_module(name)
        mod = imp.load_module(name, fp, pathname, desc)
    except ImportError:
        mod = None
    finally:
        # Since we may exit via an exception, close fp explicitly.
        if fp:
            fp.close()

    return mod

def jp_compose(s, base=None):
    """ append/encode a string to json-pointer
    """
    if s == None:
        return base

    ss = [s] if isinstance(s, six.string_types) else s
    ss = [s.replace('~', '~0').replace('/', '~1') for s in ss]
    if base:
        ss.insert(0, base)
    return '/'.join(ss)

def jp_split(s):
    """ split/decode a string from json-pointer
    """
    if s == '' or s == None:
        return []

    def _decode(s):
        s = s.replace('~1', '/')
        return s.replace('~0', '~')

    return [_decode(ss) for ss in s.split('/')]

def jr_split(s):
    """ split a json-reference into (url, json-pointer)
    """
    p = six.moves.urllib.parse.urlparse(s)
    return (
        normalize_url(six.moves.urllib.parse.urlunparse(p[:5]+('',))),
        '#'+p.fragment if p.fragment else '#'
    )

def deref(obj, guard=None):
    """ dereference $ref
    """
    cur, guard = obj, guard or CycleGuard()
    guard.update(cur)
    while cur and getattr(cur, 'ref_obj', None) != None:
        cur = cur.ref_obj
        guard.update(cur)
    return cur

def final(obj):
    return obj.final if getattr(obj, 'final', None) else obj

def get_dict_as_tuple(d):
    """ get the first item in dict,
    and return it as tuple.
    """
    for k, v in six.iteritems(d):
        return k, v
    return None

def nv_tuple_list_replace(l, v):
    """ replace a tuple in a tuple list
    """
    _found = False
    for i, x in enumerate(l):
        if x[0] == v[0]:
            l[i] = v
            _found = True

    if not _found:
        l.append(v)

def path2url(p):
    """ Return file:// URL from a filename.
    """
    # Python 3 is a bit different and does a better job.
    if sys.version_info.major >= 3 and sys.version_info.minor >= 4:
        import pathlib
        return pathlib.Path(p).as_uri()
    else:
        return six.moves.urllib.parse.urljoin(
            'file:', six.moves.urllib.request.pathname2url(p)
        )

_windows_path_prefix = re.compile(r'(^[A-Za-z]:\\)')

def normalize_url(url):
    """ Normalize url
    """
    if not url:
        return url

    matched = _windows_path_prefix.match(url)
    if matched:
        return path2url(url)

    p = six.moves.urllib.parse.urlparse(url)
    if p.scheme == '':
        if p.netloc == '' and p.path != '':
            # it should be a file path
            url = path2url(os.path.abspath(url))
        else:
            raise ValueError('url should be a http-url or file path -- ' + url)

    return url

def url_dirname(url):
    """ Return the folder containing the '.json' file
    """
    p = six.moves.urllib.parse.urlparse(url)
    for e in [private.FILE_EXT_JSON, private.FILE_EXT_YAML]:
        if p.path.endswith(e):
            return six.moves.urllib.parse.urlunparse(
                p[:2]+
                (os.path.dirname(p.path),)+
                p[3:]
            )
    return url

def url_join(url, path):
    """ url version of os.path.join
    """
    p = six.moves.urllib.parse.urlparse(url)

    t = None
    if p.path and p.path[-1] == '/':
        if path and path[0] == '/':
            path = path[1:]
        t = ''.join([p.path, path])
    else:
        t = ('' if path and path[0] == '/' else '/').join([p.path, path])

    return six.moves.urllib.parse.urlunparse(
        p[:2]+
        (t,)+ # os.sep is different on windows, don't use it here.
        p[3:]
    )

def normalize_jr(jr, url=None):
    """ normalize JSON reference, also fix
    implicit reference of JSON pointer.
    input:
    - #/definitions/User
    - http://test.com/swagger.json#/definitions/User
    output:
    - http://test.com/swagger.json#/definitions/User

    input:
    - some_folder/User.json
    output:
    - http://test.com/some_folder/User.json
    """

    if jr == None:
        return jr

    idx = jr.find('#')
    path, jp = (jr[:idx], jr[idx+1:]) if idx != -1 else (jr, None)

    if len(path) > 0:
        p = six.moves.urllib.parse.urlparse(path)
        if p.scheme == '' and url:
            p = six.moves.urllib.parse.urlparse(url)
            # it's the path of relative file
            path = six.moves.urllib.parse.urlunparse(p[:2]+('/'.join([os.path.dirname(p.path), path]),)+p[3:])
            path = derelativise_url(path)
    else:
        path = url

    if path:
        return ''.join([path, '#', jp]) if jp else path
    else:
        return '#' + jp

def _fullmatch(regex, chunk):
    m = re.match(regex, chunk)
    if m and m.span()[1] == len(chunk):
        return m

def derelativise_url(url):
    '''
    Normalizes URLs, gets rid of .. and .
    '''
    parsed = six.moves.urllib.parse.urlparse(url)
    newpath=[]
    for chunk in parsed.path[1:].split('/'):
        if chunk == '.':
            continue
        elif chunk == '..':
            # parent dir.
            newpath=newpath[:-1]
            continue
        # TODO: Verify this behaviour.
        elif _fullmatch(r'\.{3,}', chunk) is not None:
            # parent dir.
            newpath=newpath[:-1]
            continue
        newpath += [chunk]
    return six.moves.urllib.parse.urlunparse(parsed[:2]+('/'+('/'.join(newpath)),)+parsed[3:])

def get_swagger_version(obj):
    """ get swagger version from loaded json """

    if isinstance(obj, dict):
        if 'swaggerVersion' in obj:
            return obj['swaggerVersion']
        elif 'swagger' in obj:
            return obj['swagger']
        return None
    else:
        # should be an instance of BaseObj
        return obj.swaggerVersion if hasattr(obj, 'swaggerVersion') else obj.swagger

def walk(start, ofn, cyc=None):
    """ Non recursive DFS to detect cycles

    :param start: start vertex in graph
    :param ofn: function to get the list of outgoing edges of a vertex
    :param cyc: list of existing cycles, cycles are represented in a list started with minimum vertex.
    :return: cycles
    :rtype: list of lists
    """
    ctx, stk = {}, [start]
    cyc = [] if cyc == None else cyc

    while len(stk):
        top = stk[-1]

        if top not in ctx:
            ctx.update({top:list(ofn(top))})

        if len(ctx[top]):
            n = ctx[top][0]
            if n in stk:
                # cycles found,
                # normalize the representation of cycles,
                # start from the smallest vertex, ex.
                # 4 -> 5 -> 2 -> 7 -> 9 would produce
                # (2, 7, 9, 4, 5)
                nc = stk[stk.index(n):]
                ni = nc.index(min(nc))
                nc = nc[ni:] + nc[:ni] + [min(nc)]
                if nc not in cyc:
                    cyc.append(nc)

                ctx[top].pop(0)
            else:
                stk.append(n)
        else:
            ctx.pop(top)
            stk.pop()
            if len(stk):
                ctx[stk[-1]].remove(top)

    return cyc

def _diff_(src, dst, ret=None, jp=None, exclude=[], include=[]):
    """ compare 2 dict/list, return a list containing
    json-pointer indicating what's different, and what's diff exactly.

    - list length diff: (jp, length of src, length of dst)
    - dict key diff: (jp, None, None)
    - when src is dict or list, and dst is not: (jp, type(src), type(dst))
    - other: (jp, src, dst)
    """

    def _dict_(src, dst, ret, jp):
        ss, sd = set(src.keys()), set(dst.keys())
        # what's include is prior to what's exclude
        si, se = set(include or []), set(exclude or [])
        ss, sd = (ss & si, sd & si) if si else (ss, sd)
        ss, sd = (ss - se, sd - se) if se else (ss, sd)

        # added keys
        for k in sd - ss:
            ret.append((jp_compose(k, base=jp), None, None,))

        # removed keys
        for k in ss - sd:
            ret.append((jp_compose(k, base=jp), None, None,))

        # same key
        for k in ss & sd:
            _diff_(src[k], dst[k], ret, jp_compose(k, base=jp), exclude, include)

    def _list_(src, dst, ret, jp):
        if len(src) < len(dst):
            ret.append((jp, len(src), len(dst),))
        elif len(src) > len(dst):
            ret.append((jp, len(src), len(dst),))
        else:
            if len(src) == 0:
                return

            # make sure every element in list is the same
            def r(x, y):
                if type(y) != type(x):
                    raise ValueError('different type: {0}, {1}'.format(type(y).__name__, type(x).__name__))
                return x
            ts = type(functools.reduce(r, src))
            td = type(functools.reduce(r, dst))

            # when type is different
            while True:
                if issubclass(ts, six.string_types) and issubclass(td, six.string_types):
                    break
                if issubclass(ts, six.integer_types) and issubclass(td, six.integer_types):
                    break
                if ts == td:
                    break
                ret.append((jp, str(ts), str(td),))
                return

            if ts != dict:
                ss, sd = sorted(src), sorted(dst)
            else:
                # process dict without sorting
                # TODO: find a way to sort list of dict, (ooch)
                ss, sd = src, dst

            for idx, (s, d) in enumerate(zip(src, dst)):
                _diff_(s, d, ret, jp_compose(str(idx), base=jp), exclude, include)

    ret = [] if ret == None else ret
    jp = '' if jp == None else jp

    if isinstance(src, dict):
        if not isinstance(dst, dict):
            ret.append((jp, type(src).__name__, type(dst).__name__,))
        else:
            _dict_(src, dst, ret, jp)
    elif isinstance(src, list):
        if not isinstance(dst, list):
            ret.append((jp, type(src).__name__, type(dst).__name__,))
        else:
            _list_(src, dst, ret, jp)
    elif src != dst:
        ret.append((jp, src, dst,))

    return ret

def get_or_none(obj, *a):
    ret = obj
    for v in a:
        ret = getattr(ret, v, None)
        if not ret:
            break
    return ret

def patch_path(base_path, path):
    # try to get extension from base_path
    _, ext = os.path.splitext(base_path)
    if ext not in private.VALID_FILE_EXT:
        ext = ''

    # try to get extension from path
    _, ext = os.path.splitext(path) if ext == '' else (None, ext)
    if ext not in private.VALID_FILE_EXT:
        ext = ''

    # .json is default extension to try
    ext = '.json' if ext == '' else ext
    # make sure we get .json or .yaml files
    if not path.endswith(ext):
        path = path + ext

    # trim the leading slash, which is invalid on Windows
    if os.name == 'nt' and path.startswith('/'):
        path = path[1:]

    return path


class CaseInsensitiveDict(collections.abc.MutableMapping):
    """ a case insensitive dict:
        - allow to query with case insensitive keys (get, in)
        - iteration would return original key

    A reference implementation could be found in

        https://github.com/requests/
    """

    def __init__(self):
        self._store = dict()

    def __setitem__(self, key, value):
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._store[key.lower()][1]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __iter__(self):
        return (original_key for original_key, _ in six.itervalues(self._store))

    def iteritems(self):
        return six.itervalues(self._store)

    def itervalues(self):
        return (value for _, value in six.itervalues(self._store))

    def __in__(self, key):
        return key.lower() in self._store

    def __len__(self):
        return len(self._store)

    def __repr__(self):
        return str(dict(self.items()))




from __future__ import absolute_import
from ...core import BaseClient
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.httputil import url_concat, HTTPHeaders
from tornado import gen


class TornadoClient(BaseClient):
    """ Client implementation based on
    tornado.http.AsyncHTTPClient.
    """

    __schemes__ = set(['http', 'https'])

    def __init__(self, auth=None):
        """
        """
        super(TornadoClient, self).__init__(auth)
        self.__client = AsyncHTTPClient()

    @gen.coroutine
    def request(self, req_and_resp, opt=None, headers=None):
        """
        """

        # make sure all prepared state are clean before processing
        req, resp = req_and_resp
        req.reset()
        resp.reset()

        opt = opt or {}
        req, resp = super(TornadoClient, self).request((req, resp), opt)

        req.prepare(scheme=self.prepare_schemes(req), handle_files=True)
        req._patch(opt)

        composed_headers = self.compose_headers(req, headers, opt)
        tornado_headers = HTTPHeaders()
        for h in composed_headers:
            tornado_headers.add(*h)

        url = url_concat(req.url, req.query)

        rq = HTTPRequest(
            url=url,
            method=req.method.upper(),
            headers=tornado_headers,
            body=req.data
            )
        rs = yield self.__client.fetch(rq)

        resp.apply_with(
            status=rs.code,
            raw=rs.body
            )

        for k, v in rs.headers.get_all():
            resp.apply_with(header={k: v})

        raise gen.Return(resp)

